use deimos::object_file::{DvmObjectFile, DvmPath};
use deimos::vm::constant::DvmConstant;
use deimos::vm::function::DvmFunction;
use deimos::vm::instruction::{Immediate, Instruction, Location};
use deimos::vm::platform::get_std_lib_platform_functions;
use deimos::vm::source_code_location::SourceCodeLocation;
use deimos::vm::{dump_state, run_main_function, DvmContext, DvmState};
use std::rc::Rc;

fn main() {
    // Goal:
    // - write a single lib with a main()
    // - call the main fn
    //
    // ns greeter
    //
    // fn main() {
    //     println "Hello, World!"
    // }
    let mut greeter_object_file = DvmObjectFile::new(DvmPath::new("greeter", "greeter.dmo"));
    
    let main_instructions = vec![
        Instruction::MoveImmediate {
            destination: Location::Register(0),
            immediate: Immediate::Constant(DvmConstant::String(String::from("Hello, World!"))),
        },
        Instruction::Push { source_register: 0 },
        Instruction::InvokeStaticPlatform {
            function_name: Rc::new(String::from("std::core::println")),
            source_code_location: SourceCodeLocation {
                source_file_name: String::from("greeter.dm"),
                line: 4,
                char: 5,
            },
        },
        Instruction::Pop {
            // println return value
            destination_register: None,
        },
        Instruction::Pop {
            // arg0 to println
            destination_register: None,
        },
        Instruction::MoveImmediate {
            destination: Location::Register(1),
            immediate: Immediate::Int(0),
        },
        Instruction::Push { source_register: 1 },
        Instruction::Return,
    ];

    let main_function = DvmFunction::new(
        "greeter::main",
        &main_instructions,
        SourceCodeLocation {
            source_file_name: String::from("greeter.dm"),
            line: 3,
            char: 1,
        },
    );
    
    greeter_object_file.add_function(main_function);

    let mut state = DvmState::new();
    let mut context = DvmContext::new("greeter::main");
    context.add_platform_functions(get_std_lib_platform_functions());
    greeter_object_file.load_to_context(&mut context, &|path| {
        todo!()
    });
    
    let exit_code = run_main_function(&mut state, &context);

    dump_state(&state, "After main!");

    if exit_code != 0 {
        std::process::exit(exit_code);
    }
}
